import os
import json
from datetime import datetime
import time
import threading
import logging
from typing import Callable, List, Dict, Any, Union, Optional
from .typing import Numeric

def module_id_from_str(module_name: str) -> str:
    """Create module ID from module name string.

    The format is: first letter lowercase.

    Examples:
        - `Hearbeat` -> ``heartbeat``
        - `WeatherConditions` -> ``weatherConditions``

    Args:
        module_name: module name

    Returns:
        str: formatted module ID
    """
    return module_name[0].lower() + module_name[1:]

def module_id(module) -> str:
    """Create module ID from module itself

    Call :func:`module_id_from_str` on name of given module.

    Args:
        module: instance of module

    Returns:
        str: module ID is created from ``module.__name__`` attribute
    """
    return module_id_from_str(module.__name__)

def file_save(path: str, data: Union[List, Dict]) -> bool:
    """Save data to file in JSON.

    Args:
        path: file path
        data: dictionary or list with data

    Returns:
        bool: `True` if saved, `False` if file already contains the same data as passed
    """

    # prevent writing to file if not neccesarry (if power wents out at exact
    # moment, file might be empty)
    if (os.path.isfile(path) and os.stat(path).st_size > 0):  # not empty file
        with open(path, 'r') as f:
            file_json = json.load(f)
            if json.loads(json.dumps(data)) == file_json:  # force serialization on data
                return False  # do nothing, file contains the same (currently retrying)

    # save new data
    with open(path, 'w') as json_data_file:
        json_data_file.write(json.dumps(data))

    return True

def get_iso_timestamp(dt: datetime) -> str:
    """Create ISO timetamp from given datetime object.

    Args:
        dt: :class:`datetime.datetime` object

    Returns:
        str: corresponding ISO timestamp
    """
    return dt.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'

def get_datetime_from_iso_timestamp(iso_timestamp: str) -> datetime:
    """Create datetime object from given ISO timetamp .

    Args:
        iso_timestamp: ISO timestamp

    Returns:
        datetime: corresponding :class:`datetime.datetime` object
    """
    return datetime.strptime(iso_timestamp, '%a, %d %b %Y %H:%M:%S %Z')

class LoggingSystemdHandler(logging.StreamHandler):
    """ Severity information for stdout&stderr logging. See SD-DAEMON(3) """
    PREFIX = {
        # EMERG <0>
        # ALERT <1>
        logging.CRITICAL: "<2>",
        logging.ERROR: "<3>",
        logging.WARNING: "<4>",
        # NOTICE <5>
        logging.INFO: "<6>",
        logging.DEBUG: "<7>",
        logging.NOTSET: "<7>"
    }
    def emit(self, record):
        try:
            msg = self.PREFIX[record.levelno] + self.format(record)
            msg = msg.replace("\n", "\\n") # Tracebacks on single line
            self.stream.write(msg + "\n")
            self.stream.flush()
        except Exception:
            self.handleError(record)

logger_rt = logging.getLogger(__name__ + ".RepeatTimer")
class RepeatTimer:
    """ Periodically run function with unlimited repetition.

    Start the timer by executing :meth:`.start`. The timer can be interrupted
    at any time by issuing :meth:`.stop`.

    It is possible to :meth:`.stop` the timer, directly set different :attr:`period` or
    :attr:`runonstart` and :meth:`.start` again. But be sure to STOP before changing
    these values. Any exceptions raised in :attr:`f` will be logged with ERROR log
    level using `logging` library together with stacktrace. This will NOT stop
    the timer.

    Warning:
        Any exception occuring in the function :attr:`f` is catched and logged,
        ie. exceptions occuring in the function will not stop the timer.

    Args:
        period: period between timer ticks, in seconds
        f: function to be executed
        args: positional arguments to be passed to ``f``
        kwargs: keyword arguments to be passed to ``f``
        runonstart: default to `True` -- make the first execution of ``f`` right after
            calling :meth:`.start` (without waiting first for ``period``)
    """

    _period: Numeric
    _runonstart: bool
    _f: Callable
    _args: List
    _kwargs: Dict
    _stopev: threading.Event

    def __init__(self,
            period: Numeric, f: Callable,
            args: List=[], kwargs: Dict={},
            runonstart: bool=True
        ):
        self._period = period
        self._runonstart = runonstart

        self._f = f
        self._args = args
        self._kwargs = kwargs

        # stop event to abort wait (would be time.sleep) in while loop
        self._stopev = threading.Event()
        self._stopev.set()

    @property
    def period(self) -> Numeric:
        return self._period
    @period.setter
    def period(self, new) -> None:
        if not self._stopev.is_set():
            raise RuntimeError("Stop the timer before setting `period`.")
        self._period = new

    @property
    def runonstart(self) -> bool:
        return self._runonstart
    @runonstart.setter
    def runonstart(self, new) -> None:
        if not self._stopev.is_set():
            raise RuntimeError("Stop the timer before setting `runonstart`.")
        self._runonstart = new

    def start(self):
        """ Start the timer. """
        # make timer possible to restart again
        self._stopev.clear()
        self._t = threading.Thread(target=self._do_every)
        self._t.start()

    def stop(self, block: bool=True):
        """Stop the timer.

        Can be set to block until current execution of :attr:`.f` is finished.

        Args:
            block: If ``block`` is `True`, block until currently executed
                function is finished. Default to `True`.
        """
        self._stopev.set()
        # race conditions
        if block and self._t.is_alive():
            self._t.join()

    def _do_every(self):
        def sleep_sequence():
            t = time.time()
            while True:
                t += self._period
                yield max(t - time.time(), 0)

        s = sleep_sequence()

        # first wait
        if not self._runonstart:
            self._stopev.wait(next(s))

        # while stop was not called
        while not self._stopev.is_set():
            try:
                self._f(*self._args, **self._kwargs)
            except Exception as e:
                logger_rt.exception("Function [%s] thrown an error: [%s]. See stacktrace. ",
                        self._f, str(e))

            # wait for stopevent for next(s) seconds
            self._stopev.wait(next(s))

logger_st = logging.getLogger(__name__ + ".Storage")
class Storage:
    """Storage providing thread-safe utility functionality.

    This class provides:

        - basic thread-safe storage centered around Python's ``list``
        - backup and restore the storage to JSON files
        - limiting the storage to a number of entries. If enabled, the oldest
          entries are discarded. Newest entries are the ones stored at the end
          of the list.

            - as the limited storages are typically small in size (~100), no
              optimization is done for the sake of code simplicity. The
              ``list``'s removal from the left is ``O(n)`` where ``n`` is the
              size of the storage. If you need bigger limited storages,
              consider double ended queue ``deque``.

    Args:
        storage_id: identifier used only for logging purposes
        backup_path: location of backup JSON file. Pass ``None`` to turn off
            file backups. It will automatically create empty directories and
            create the backup file, if the path does not exist. If it exists,
            load data into storage from file. 
        maxlen: maximum number of entries in the storage. Oldest entries are
            replaced by newer ones. Pass ``0`` to make the storage unlimited.
    """

    #: writing lock :class:`threading.Lock` can be used for more complex
    #: operations outside this class and is public
    write_lock: threading.Lock

    #: storage id, used for logging
    _sid: str
    #: max length of the storage
    _maxlen: int
    _data: List
    _backup_path: Optional[str]

    def __init__(self, storage_id: str, backup_path: Optional[str], maxlen: int=0):
        self.write_lock = threading.Lock()

        self._sid = storage_id
        self._maxlen = maxlen

        self._data = []

        self._backup_path = backup_path
        if not self._backup_path:
            return

        os.makedirs(os.path.dirname(self._backup_path), exist_ok=True)
        if os.path.exists(self._backup_path):
            with open(self._backup_path, 'r') as f:
                try:
                    data = json.load(f)
                except json.decoder.JSONDecodeError:
                    data = []
                    logger_st.warning("Backup file %s is not a valid JSON file.",
                            self._backup_path)
                    return

            if data:
                self._data = data
                logger_st.debug("Backup file `%s` for `%s` loaded successfully with [%i] entries",
                        self._backup_path, self._sid, len(data))
            else:
                logger_st.debug("Nothing to load from backup file `%s` for `%s`",
                        self._backup_path, self._sid)

        else:
            with open(self._backup_path, 'w') as f:
                f.write('[]')
            logger_st.debug("Created new backup file for `%s`", self._sid)

    def __len__(self):
        return len(self._data)

    @property
    def data(self) -> List:
        """ The stored data list. """
        return self._data

    def append(self, data: Any, backup: bool=False) -> None:
        """ Append a piece of new data to the storage list. Thread-safe.

        There is an option to backup the storage directly after the append.
        Default is to not back up to file, because of the unnecessary IO
        overhead.

        Args:
            data: data to be appended using :meth:`self.data.append`
            backup: backup the storage to the file after appending. Defaults to ``False``.
        """
        with self.write_lock:
            self._data.append(data)
            self.trim()
            if backup:
                self._save_to_disk()

    def merge(self, data: List, new_data: bool=True, backup: bool=True) -> None:
        """ Merge another list with the storage. Thread-safe.

        Chronologically merge data with storage. Defaults to file backup after
        the operation, if enabled in constructor.

        Args:
            data: another list to merge with the data.
            new_data: Value ``True`` means ``data`` are newer than whole
                content of the storage and are stored at the end of the storage.
                This is assumed by default. Value ``False`` merges the data at the
                beginning of the storage. This is important when storage limit is
                imposed, as the oldest entries will be deleted when the total number
                of entries is greater that the limit.
            backup: backup the storage to the file after appending. Defaults to ``True``.

        """
        with self.write_lock:
            if new_data:
                self._data = self._data + data
            else:
                self._data = data + self._data
            self.trim()
            if backup:
                self._save_to_disk()

    def is_empty(self) -> bool:
        """ Is the storage an empty list ``[]``? """
        return self._data == []

    def empty(self) -> None:
        """ Empty the storage and empty the backup file. """
        logger_st.debug("Emptying storage `%s`.", self._sid)
        self._data = []
        self._save_to_disk()

    def trim(self) -> None:
        """ Trim the storage size to the limit ``limit``.

        Only ``limit`` last entries of the storage are kept, rest is deleted.

        Do not trim if the ``limit`` is not set (is zero).
        """
        # equivalent even for edge cases in condition below
        # slen = len(self)
        # if self._limit > 0 and slen > self._limit:
            # del self._storage[:-self._limit]

        # keep last self._limit newest entries
        del self._data[:-self._maxlen]

    def _save_to_disk(self):
        """ Save the storage to disk.

        Saves the storage to backup path, if specified. Otherwise, do nothing.
        """
        if self._backup_path:
            if file_save(self._backup_path, self._data):
                logger_st.debug("Saving storage of `%s` to backup file.", self._sid)
            else:
                logger_st.debug("Storage already saved, doing nothing.")



# Exception for specifying unknown event in `on()` function
class UnsupportedEvent(Exception):
    """Unknown gateway event."""


class UuAppClientException(Exception):
    """Internal error in request pre-processing occured."""


class TokenError(UuAppClientException):
    """Error when getting/validating a token occured."""


class TokenCommandError(TokenError):
    """Server error when getting a token occured."""



"""Initialize modules."""
from typing import Dict
import logging
from uun_iot.UuAppClient import UuAppClient
from .Heartbeat import Heartbeat
from .BaseHealthCheck import BaseHealthCheck

def init(config: Dict, uuclient: UuAppClient):

    def cmd_heartbeat(dto_in):
        uucmd = config["uuApp"]['uuCmdList']['gatewayHeartbeat']
        resp, exc = uuclient.post(uucmd, dto_in, log_level=logging.DEBUG)
        if exc is not None:
            return False
        return resp

    return [Heartbeat(cmd_heartbeat)]



""" Communication with the *.Main uuSubApp. """

import datetime
import json
import logging
import time
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder
from requests_toolbelt.utils import user_agent as ua

import uun_iot

from .exceptions import TokenCommandError, TokenError, UuAppClientException

logger = logging.getLogger(__name__)


class UuAppClient:
    """Library functions for communication and authentication with uuApp.

    The class gets a dictionary ``config`` containing the configuration of the application.
    The (root) keys and subkeys of interest to this class are

        - ``uuApp``: information about server uuApp

            - ``gateway``: domain gateway of the corresponding uuApp
            - ``uuAppName``: full name of the uuApp
            - ``awid``
            - ``uuCmdList``: containing uuCmd endpoint. The keys are not used
              by :class:`UuAppClient` class - they are used by the application
              user modules (most commonly specified in Python package path
              ``<package_name>/modules/__init__.py``.

        - ``uuThing``: authentication information about the IoT gateway

            - ``uuIdentity``
            - ``accessCode1``
            - ``accessCode2``

        - ``oidcGrantToken``: information about the authentication token
          service used in server communication

            - ``gateway``: (usually ``uuidentity.plus4u.net``) domain gateway of the `OIDC` service
            - ``uuAppName``: (usually ``uu-oidc-maing02``) full name of the `OIDC` uuApp
            - ``awid``
            - ``clientId``
            - ``clientSecret``
            - ``uuCmd``: (usually ``oidc/grantToken``) uuCmd enpoint for token granting
            - ``tokenPath``: (usually ``./oidc-token``) filesystem location (on
              the IoT gateway) where the token will be saved

    Args:
        config: dictionary with configuration, see above
        fn_post_request: optional function to get a token response from server,
            defaults to :meth:`requests.post`
        refresh_token_on_init: call :meth:`.refresh_token` on initialization of this class
        token_leeway: leeway (drift) between Client and Server Clock [s], defaults to 60 s
            fresh token is requested when token is to be expired in less than token_leeway [s]
            the lower the leeway, the more often is token going to be requested
            the token validity time is controlled by the server, this option accounts only for drift
    """

    _config: dict
    _token: Optional[str]
    _user_agent: str
    _token_expires_at: Optional[float]
    _auth_active: bool
    _token_leeway: float

    def __init__(
        self,
        config: dict,
        fn_token_request: Callable[..., requests.Response] = requests.post,
        refresh_token_on_init: bool = True,
        token_leeway: float = 60,
    ):
        self._config = config
        self._token = None
        self._token_expires_at = None
        self._user_agent = ua.UserAgentBuilder("uun-iot", uun_iot.__version__).build()
        self._token_leeway = token_leeway
        self._auth_active = (
            "uuApp" in self._config
            and "uuThing" in self._config
            and "oidcGrantToken" in self._config
        )
        self._fn_token_request = fn_token_request

        if refresh_token_on_init and self._auth_active:
            self.refresh_token()
        else:
            if refresh_token_on_init:
                logger.info(
                    "UuApp connection & gateway authentication is not "
                    "defined in the configuration file. Are you sure, you do not "
                    "want to use the connection? It will not be possible to "
                    "issue requests to uu servers."
                )
            logger.debug("Not granting any auth token.")

    def _store_token(self, token_json: dict) -> None:
        """Save serialized token to file."""
        with open(self._config["oidcGrantToken"]["tokenPath"], "w") as file:
            file.write(json.dumps(token_json))

    def is_token_expired(self) -> bool:
        """
        # view with respect to server-centered time
        #       <-> leeway
        # ++++++==|-----
        #         |
        #    server_token_expire
        #
        # +: when (local) now is there, token is valid
        # =: when (local) now is there, token is valid but apply for a new token
        #      because of the possible leeway between client and server clocks
        # -: when (local) now is there, token is invalid
        """
        return (
            self._token_expires_at is None
            or self._token_expires_at - self._token_leeway < time.time()
        )

    def refresh_token(self) -> bool:
        """Refresh token if expired. If authentication is disabled, noops and returns False.

        Returns False if token refresh was not needed - that is, if token is
        already loaded and valid. Refresh token and return True otherwise.

        Raises:
            TokenError: when a valid token could not be obtained due to network or server error
            TokenCommandError (subclass of TokenError): server-side uuApp returned error
        """
        if not self._auth_active:
            return False

        if self._token is None:
            # load from file
            token_path = self._config["oidcGrantToken"]["tokenPath"]
            try:
                with open(token_path) as json_file:
                    token = json.load(json_file)
                    self._token = token["id_token"]
                    self._token_expires_at = float(token["expires_at"])

                logger.debug(
                    "loaded token from %s, expires at [%s]",
                    token_path,
                    datetime.datetime.fromtimestamp(self._token_expires_at),
                )
            except IOError:
                logger.debug("loading token from %s failed, file not found", token_path)
            except json.JSONDecodeError:
                logger.debug("Invalid JSON token file.")
            except KeyError:
                logger.debug("Invalid structure of token file.")

        if self.is_token_expired():
            self._token_expires_at = self._grant_token(self._fn_token_request)
            logger.debug(
                "granted new token, expires at [%s]",
                datetime.datetime.fromtimestamp(self._token_expires_at),
            )
            return True

        return False

    def _grant_token(self, fn_post_request: Callable[..., requests.Response]) -> float:
        """Grant token.

        Refresh and save the new token.

        Args:
            fn_post_request: function to get a token response from server

        Returns:
             expiry date as unix timestamp

        Raises:
            TokenError: when a valid token could not be obtained due to network or server error
            TokenCommandError: server-side uuApp returned error
        """
        scope_host = self._config["uuApp"]["gateway"]
        scope_context = self._config["uuApp"]["uuAppName"]
        scope_awid = self._config["uuApp"]["awid"]
        post_data = {
            "grant_type": "password",
            "username": None,
            "password": None,
            "accessCode1": self._config["uuThing"]["accessCode1"],
            "accessCode2": self._config["uuThing"]["accessCode2"],
            "scope": f"openid https://{scope_host}/{scope_context}/{scope_awid}",
        }

        try:
            post_data.update(
                {
                    "client_id": self._config["oidcGrantToken"]["clientId"],
                    "client_secret": self._config["oidcGrantToken"]["clientSecret"],
                }
            )
        except KeyError:
            # does not matter if these are not present - not required - depends on server settings
            pass

        headers = {"Content-Type": "application/json", "User-Agent": self._user_agent}
        uucmd = self._config["oidcGrantToken"]["uuCmd"]

        host = self._config["oidcGrantToken"]["gateway"]
        uu_app_name = self._config["oidcGrantToken"]["uuAppName"]
        awid = self._config["oidcGrantToken"]["awid"]
        url = f"https://{host}/{uu_app_name}/{awid}/{uucmd}"
        response = None
        try:
            response = fn_post_request(url, headers=headers, json=post_data, timeout=20)
            response.raise_for_status()
        except requests.HTTPError as e:
            assert response is not None
            logger.log(
                logging.DEBUG,
                "`%s` POST response (%s): %s",
                uucmd,
                response.status_code,
                response.text,
            )
            status = response.status_code
            try:
                error_json = response.json()
            except json.JSONDecodeError as exc:
                raise TokenError(
                    "A server error occured while getting a token and could not decode"
                    " JSON error object."
                ) from exc
            if "uuAppErrorMap" in error_json:
                raise TokenCommandError(error_json["uuAppErrorMap"], status) from e
            raise TokenError(
                "Server error occured but could not get more information."
            ) from e
        except requests.RequestException as e:
            logger.log(logging.WARNING, "`%s` POST RequestException: %s", uucmd, str(e))
            raise TokenError(
                "A network error occured when requesting new token."
            ) from e

        try:
            token_json = response.json()
            self._token = token_json["id_token"]
            token_json["expires_at"] = time.time() + float(token_json["expires_in"])
        except json.JSONDecodeError as e:
            raise TokenError(
                "While getting a token, could not decode received JSON server response."
            ) from e
        except KeyError as e:
            raise TokenError("Token did not have required fields.") from e
        except ValueError as e:
            raise TokenError("Token has invalid value types.") from e

        self._store_token(token_json)
        return token_json["expires_at"]

    def get_uucmd_url(self, uucmd: str) -> str:
        """Return fully quallified URL from relative uucmd string.

        Args:
            uucmd: uucmd in relative path, eg. ``gateway/getWeather``
        """
        host = self._config["uuApp"]["gateway"]
        uu_app_name = self._config["uuApp"]["uuAppName"]
        awid = self._config["uuApp"]["awid"]
        return f"https://{host}/{uu_app_name}/{awid}/{uucmd}"

    def get_auth_headers(self, content_type: str = "application/json"):
        """Get request headers with Bearer authentication.
        The token is refreshed automatically, if expired.

        Headers also includes specified ``Content-type`` and uun-iot's ``User-Agent``.
        """
        if self.is_token_expired():
            self.refresh_token()
        return {
            "Authorization": f"Bearer {self._token}",
            "Content-type": content_type,
            "User-Agent": self._user_agent,
        }

    def custom_request(
        self,
        custom_send_f: Callable[[object], requests.Response],
        dto_in: Optional[object] = None,
        label: Optional[str] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """
        Custom request wrapper for catching :mod:`requests`-related and token-acquiring-related
        Exceptions. Token errors, request exceptions, including HTTP non-success status codes,
        are suppressed, logged with specified level and returned.

        Supply your own :attr:`custom_send_f` function (taking :attr:`dto_in` as
        only argument) and outputting a :attr:`requests.Response`. To make an
        authenticated request, you should use authentication headers available
        through :meth:`.get_auth_headers`. Note that you can specify the headers
        either directly during *definition*, or during *evaluation* of
        :attr:`custom_send_f`. The latter is preferred due to handling TokenErrors
        in a unified way with other exceptions. Note that token acquisition is not in
        scope of :meth:`.custom_request` and has to be supplied by caller.

        See also :meth:`.get_uucmd_url` and :meth:`.get_auth_headers` for
        constructing own function.

        .. note::
            Example for ``custom_send_f`` follows.

            .. code-block:: python

                uucmd = "gateway/heartbeat"
                full_url = uuappclient.get_uucmd_url(uucmd)
                def handler(dto_in):
                    return requests.get(full_url, headers=self.get_auth_headers(), json=dto_in_, timeout=20)
                uuappclient.custom_request(handler, {"key": "value"}, uucmd)


        This function will
            - call :attr:`custom_send_f` with :attr:`dto_in` positional
              argument,
            - raise for non-success HTTP status codes
            - catch :attr:`requests.HTTPError` and
              :attr:`requests.RequestException`,
            - log the exceptions and
            - returns response along with the exception

        Args:
            custom_send_f: custom send function which is used instead of
                :meth:`requests.get` label: label used in log messages to identify the request
            dto_in: dictionary data to send. If ``None``, pass ``{}`` to :attr:`custom_send_f`.
            log_level: optional logging level for possible exceptions, defaults to ``logging.WARNING``

        Returns: Tuple ``(response, exc)``,
                - :class:`requests.Response` response, ``None`` if non-HTTP (not
                  200 HTTP status code) exception was raised
                - ``exc``: :class:`requests.RequestException` or :class:`TokenError` if the same
                  exception was raised during request. ``None`` if no exception during request was raised

        Raises: None
        """

        if dto_in is None:
            dto_in = {}

        response = None
        exc = None
        try:
            response = custom_send_f(dto_in)
            response.raise_for_status()
        except requests.HTTPError as e:
            assert response is not None
            logger.log(
                log_level,
                "`%s` response (%s): %s",
                label,
                response.status_code,
                response.text,
            )
            exc = e
        except requests.RequestException as e:
            logger.log(log_level, "`%s` RequestException: %s", label, str(e))
            exc = e
        except TokenError as e:
            logger.log(log_level, "`%s` TokenError: %s", label, str(e))
            exc = e

        return response, exc

    def get(
        self,
        uucmd: str,
        dto_in: Optional[dict] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """
        Get request using an authenticated Bearer token.
        Token errors, request exceptions, including HTTP non-success status codes, are suppressed,
        logged with specified level and returned.
        See :meth:`.custom_request` for more informaton.

        Args:
            uucmd: UuCmd in relative path format, eg. ``gateway/heartbeat``
            dto_in: data to send
            log_level: optional logging level for possible exceptions,
                defaults to ``logging.WARNING``

        Returns:
            Tuple ``(response, exc)``,
                - :class:`requests.Response` response, ``None`` if non-HTTP (not 200 HTTP status code)
                  exception was raised
                - :class:`requests.RequestException` exc is None if no exception
                  during request was raised
        """

        full_url = self.get_uucmd_url(uucmd)

        def get_f(dto_in_):
            # ask for headers in the handler function, custom_request first refreshes token
            logger.debug("`%s` GET request, payload: %s", uucmd, dto_in)
            return requests.get(
                full_url, headers=self.get_auth_headers(), json=dto_in_, timeout=20
            )

        return self.custom_request(get_f, dto_in, uucmd, log_level)

    def get_ignore_http_err(
        self,
        uucmd: str,
        dto_in: Optional[dict] = None,
        log_level: int = logging.WARNING,
    ) -> requests.Response:
        """
        GET request with authentication. Ignore HTTP errors, raise errors for other network errors.
        Logs all errors (including HTTP non-200) with :attr:`log_level` severity.
        See :meth:`.custom_request` for more informaton.

        Args:
            uucmd: uuCmd path of the target uuApp
            dto_in: data to be passed as JSON input to the uuApp
            log_level: logging library level of network error, default logging.WARNING

        Raises: :class:`requests.RequestException` or :class:`TokenError`
        """

        r, exc = self.get(uucmd, dto_in, log_level)
        if exc is not None:
            try:
                raise exc
            except requests.HTTPError:
                pass
        assert r is not None
        return r

    def post(
        self,
        uucmd: str,
        dto_in: Optional[Union[dict, MultipartEncoder]] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """
        POST request using an authenticated Bearer token.
        Request exceptions, including HTTP non-success status codes, are suppressed, logged with specified level and returned.
        See :meth:`.custom_request` for more informaton.

        Args:
            uucmd: UuCmd in relative path format, eg. ``gateway/heartbeat``
            dto_in: data to send, dictionary or MultipartEncoder data
            log_level: optional logging level for possible exceptions,
                defaults to ``logging.WARNING``

        Returns:
            Tuple ``(response, exc)``,
                - :class:`requests.Response` response, ``None`` if non-HTTP (not 200 HTTP status code)
                  exception was raised
                - :class:`requests.RequestException` exc is None if no exception
                  during request was raised

        Raises: ValueError if ``dto_in`` type is not ``dict`` or ``MultipartEncoder``
        """

        if dto_in is None:
            dto_in = {}

        full_url = self.get_uucmd_url(uucmd)
        if isinstance(dto_in, dict):
            headers = self.get_auth_headers
            def post_f(dto_in_):
                logger.debug("`%s` POST request, payload: %s", uucmd, dto_in)
                return requests.post(full_url, headers=headers(), json=dto_in_, timeout=20)
        elif isinstance(dto_in, MultipartEncoder):
            headers = partial(self.get_auth_headers, dto_in.content_type)
            def post_f(dto_in_):
                logger.debug("`%s` MULTIPART request, payload: %s", uucmd, dto_in)
                return requests.post(full_url, headers=headers(), data=dto_in_, timeout=20)
        else:
            raise ValueError(f"Invalid dto_in type: {type(dto_in)}")

        return self.custom_request(post_f, dto_in, uucmd, log_level)

    def post_ignore_http_err(
        self,
        uucmd: str,
        dto_in: Optional[Union[dict, MultipartEncoder]] = None,
        log_level: int = logging.WARNING,
    ) -> requests.Response:
        """
        POST request with authentication. Ignore HTTP errors, raise errors for other network errors.
        Logs all errors (including HTTP non-200) with :attr:`log_level` severity.

        Args:
            uucmd: uuCmd path of the target uuApp
            dto_in: data to be passed as JSON input to the uuApp
            log_level: logging library level of network error, default logging.WARNING

        Raises: :class:`requests.RequestException` or :class:`TokenError`
        """
        r, exc = self.post(uucmd, dto_in, log_level)
        if exc is not None:
            try:
                raise exc
            except requests.HTTPError:
                pass
        assert r is not None
        return r

    def multipart(
        self,
        uucmd: str,
        dto_in: Optional[dict] = None,
        log_level: int = logging.WARNING,
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """POST request with authentication and MULTIPART encoded data
        with oidc2 authentication. Useful for sending binary data (images, ...). See
        https://toolbelt.readthedocs.io/en/latest/user.html#multipart-form-data-encoder
        for information about multipart encoder.

        Pass data to dto_in, they will be transformed using MultipartEncoder and passed to
        :meth:`UuAppClient.post` (see for more information and usage).
        """

        if dto_in is None:
            return self.post(uucmd, None, log_level)
        multipart_encoded = MultipartEncoder(fields=dto_in)
        return self.post(uucmd, multipart_encoded, log_level)


class UuCmdSession:
    """
    Send all data with UuCmd in one session to avoid creating multiple connections.

    Args:
        uuclient: initialized authentication client
        uucmd: UuCmd in relative path format, eg. ``gateway/getWeather``
        log_level: ``logging`` library log level to use for network and HTTP errors,
            defaults to :attr:`logging.WARNING`
    """

    def __init__(
        self, uuclient: UuAppClient, uucmd: str, log_level: int = logging.WARNING
    ):
        self._uucmd = uucmd
        self._loglevel = log_level
        self._url = uuclient.get_uucmd_url(uucmd)
        self._session = requests.Session()
        self._uuclient = uuclient

    def update_headers(self):
        """Update session auth headers ensured to containing a valid Bearer token."""
        self._session.headers.update(self._uuclient.get_auth_headers())

    def get(
        self, data: Optional[dict] = None, log_level: int = logging.WARNING
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """Authenticated GET request with session.
        See :meth:`UuAppClient.get` for more information."""

        full_url = self._uuclient.get_uucmd_url(self._uucmd)

        def get_f(dto_in_):
            self.update_headers()
            logger.debug("`%s` GET request, payload: %s", self._uucmd, data)
            return self._session.get(full_url, json=dto_in_, timeout=20)

        return self._uuclient.custom_request(get_f, data, self._uucmd, log_level)

    def post(
        self, data: Optional[dict] = None, log_level: int = logging.WARNING
    ) -> Tuple[
        Optional[requests.Response],
        Optional[Union[TokenError, requests.RequestException]],
    ]:
        """Authenticated POST request with session.
        See :meth:`UuAppClient.post` for more information."""

        full_url = self._uuclient.get_uucmd_url(self._uucmd)

        def post_f(dto_in_):
            self.update_headers()
            logger.debug("`%s` POST request, payload: %s", self._uucmd, data)
            return self._session.post(full_url, json=dto_in_, timeout=20)

        return self._uuclient.custom_request(post_f, data, self._uucmd, log_level)



import setuptools
import os
import re

lib_path = os.path.abspath(os.path.dirname(__file__))
with open(f"{lib_path}/uun_iot/__init__.py", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

setuptools.setup(
    name="uun-iot",
    version=version,
    author="(UUN) Tomáš Faikl",
    author_email="tomas.faikl@unicornuniversity.net",
    description="Modular framework for communication with UuApp and utility functionality",
    url="https://uuapp.plus4u.net/uu-bookkit-maing01/38c7532545984b3797c5719390b523a8/book/page?code=71150832",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    install_requires=[
        "requests>=2.24.0",
        "requests_toolbelt>=0.9.1",
        "psutil>=5.7.0"
    ],
    python_requires='>=3.6'
)
